//+------------------------------------------------------------------+
//|                                                      ZigZags.mqh |
//|                                         Copyright 2011, AlexSTAL |
//|                                           http://www.alexstal.ru |
//+------------------------------------------------------------------+
#property copyright "Copyright 2011, AlexSTAL"
#property link      "http://www.alexstal.ru"

#include <CustPrevCalculated.mqh>
#include <OutsideBar.mqh>

//+------------------------------------------------------------------+
//|     ZZ                           |
//+------------------------------------------------------------------+
enum CPCZZResultCode
  {
   CPCZZRC_NotInitialized,             //   
   CPCZZRC_NoData,                     //     (      )
   CPCZZRC_NotChanged,                 //   ZZ  
   CPCZZRC_Changed                     //   ZZ
  };

//+------------------------------------------------------------------+
//|     ZZ                              |
//+------------------------------------------------------------------+
struct ZZBar
  {
   double UP, DN;                      //   ,   
   OrderFormationBarHighLow OB;        //     
  };

//+------------------------------------------------------------------+
//| Class CCustZigZagPPC - -   ZigZag PPC |
//| http://www.mql5.com/ru/code/263                                  |
//+------------------------------------------------------------------+
class CCustZigZagPPC
  {
protected:
   bool Initialized;                         //         
   bool UsesExternalData;                    //   
   //  
   //----------------------------
   int BarsLimit;                            //    ,     
   double iMinMotion;                        //           
   bool iUseSmallerTFforEB;                  //        High/Low 
   //    
   //----------------------------
   double LastBarLastHigh;                   //     
   double LastBarLastLow;                    //     
   uint DataBarsCount;                       //        ,     BarsLimit
   uint DataBarsCalculated;                  //     
   uint DataStartInd;                        //        (  )
   ZZBar Buf[];                              //     
   uint ZZDataBarsCalculated;                //     ZZZ
   //----------------------------
   uchar iExtPeriod;                         //     ,     2
   double iMinAmplitude;                     //         (  )
   datetime TimeFirstExtBar;                 //      PPC (      )
   char _Comb(int i, double H, double L, double Fup, double Fdn);
   int _iHighest(int startPos);
   int _iLowest(int startPos);
   bool ZZDataBarsCuttingLeft;               //         PrepareData()
   CCustPrevCalculated *CustPrevCalculated;
   //  
   //----------------------------
   int __Digits;
   double __Point;
   uint PInd(uint I, bool AsSeries) {return( AsSeries?((DataStartInd+DataBarsCount-1-I) % BarsLimit):((DataStartInd+I) % BarsLimit) );}
public:
   //   
   void CCustZigZagPPC() {Initialized = false;}
   void ~CCustZigZagPPC() {if (CustPrevCalculated != NULL) delete CustPrevCalculated;}
   bool ItIsInitialized() {return(Initialized);}
   //   
   bool Init(CCustPrevCalculated *CPC, string Instr, ENUM_TIMEFRAMES TF, int Limit, CPCHistorySynchMode HSM, uchar HS, uint HSMinute, uchar MinMotion, bool UseSmallerTFforEB, uchar ExtPeriod, uchar MinAmplitude);
   //    
   CPCPrepareDataResultCode PrepareData(CPCPrepareDataResultCode resData);   //       
   //    
   CPCZZResultCode Calculate();
   //     
   uint GetBarsCount() {return(DataBarsCount);}                //    
   uint GetBarsCalculated() {return(ZZDataBarsCalculated);}    //    
   double GetUP(uint shift, bool AsSeries);                    //    ZigZag  
   double GetDN(uint shift, bool AsSeries);                    //    ZigZag  
   OrderFormationBarHighLow GetOB(uint shift, bool AsSeries);  //   Outside   
  };
//+------------------------------------------------------------------+
bool CCustZigZagPPC::Init(CCustPrevCalculated *CPC, string Instr, ENUM_TIMEFRAMES TF, int Limit, CPCHistorySynchMode HSM, uchar HS, uint HSMinute, uchar MinMotion, bool UseSmallerTFforEB, uchar ExtPeriod, uchar MinAmplitude)
{
   if (Initialized)
      return(false);
   
   if (CPC == NULL)
     {
      CustPrevCalculated = new CCustPrevCalculated;
      UsesExternalData = false;
     } else {
      CustPrevCalculated = CPC;
      UsesExternalData = true;
     }
   if (CustPrevCalculated.InitData(Instr, TF, Limit, HSM, HS, HSMinute))
     {
      if (ArrayResize(Buf, Limit) == -1)
         return(Initialized);

      BarsLimit = Limit;
      
      __Digits = (int)SymbolInfoInteger(Instr, SYMBOL_DIGITS);
      __Point = SymbolInfoDouble(Instr, SYMBOL_POINT);
      
      if (ExtPeriod >= 2)
         iExtPeriod = ExtPeriod;
      else
         iExtPeriod = 2;
      
      iMinAmplitude = NormalizeDouble(MinAmplitude * __Point, __Digits);
      
      if (MinMotion >= 1)
         iMinMotion = NormalizeDouble(MinMotion * __Point, __Digits);
      else
         iMinMotion = NormalizeDouble(__Point, __Digits);
      
      iUseSmallerTFforEB = UseSmallerTFforEB;
      
      Initialized = true;
      return(true);
     } else {
      Initialized = false;
      return(false);
     }
}
//+------------------------------------------------------------------+
CPCPrepareDataResultCode CCustZigZagPPC::PrepareData(CPCPrepareDataResultCode resData)
{
   if (!Initialized)
      return(CPCPDRC_NoData);
   
   //     
   CPCPrepareDataResultCode res = resData;
   DataBarsCount = CustPrevCalculated.GetDataBarsCount();
   DataBarsCalculated = CustPrevCalculated.GetDataBarsCalculated();
   DataStartInd = CustPrevCalculated.GetDataStartInd();
   
   //     ,      
   //   ,     
   if (CustPrevCalculated.GetDataBarsCuttingLeft())
      ZZDataBarsCuttingLeft = true;
   
   //         ZZ
   for (uint i = (DataBarsCalculated == 0)?0:(DataBarsCalculated+1); i < DataBarsCount; i++)
     {
      Buf[PInd(i, false)].UP = EMPTY_VALUE;
      Buf[PInd(i, false)].DN = EMPTY_VALUE;
      Buf[PInd(i, false)].OB = OFBError;
     }
   
   //     ,     ZZ
   ZZDataBarsCalculated = MathMin(DataBarsCalculated, ZZDataBarsCalculated);
   
   //       
   double LastHigh = CustPrevCalculated.GetDataHigh(DataBarsCount-1, false);
   double LastLow = CustPrevCalculated.GetDataLow(DataBarsCount-1, false);
   if (res == CPCPDRC_SynchOnlyLastBar)
     {
      if ( (NormalizeDouble(LastHigh-LastBarLastHigh, __Digits) >= iMinMotion) || (NormalizeDouble(LastBarLastLow-LastLow, __Digits) >= iMinMotion) )
        {
         LastBarLastHigh = LastHigh;
         LastBarLastLow = LastLow;
        } else {
         //       ,     
         //           ,
         //    MinMotion
         res = CPCPDRC_NoRecountNotRequired;
         DataBarsCalculated = DataBarsCalculated + 1;
        }
     } else {
      LastBarLastHigh = LastHigh;
      LastBarLastLow = LastLow;
     }
   
   return(res);
}
//+------------------------------------------------------------------+
double CCustZigZagPPC::GetUP(uint shift, bool AsSeries)
{
   if ( shift > (DataBarsCount-1) )
      return(EMPTY_VALUE);
   return(Buf[PInd(shift, AsSeries)].UP);
}
//+------------------------------------------------------------------+
double CCustZigZagPPC::GetDN(uint shift, bool AsSeries)
{
   if ( shift > (DataBarsCount-1) )
      return(EMPTY_VALUE);
   return(Buf[PInd(shift, AsSeries)].DN);
}
//+------------------------------------------------------------------+
OrderFormationBarHighLow CCustZigZagPPC::GetOB(uint shift, bool AsSeries)
{
   if ( shift > (DataBarsCount-1) )
      return(0);
   return(Buf[PInd(shift, AsSeries)].OB);
}
//+------------------------------------------------------------------+
//|  searching index of the highest bar                              |
//+------------------------------------------------------------------+
int CCustZigZagPPC::_iHighest(int startPos)
  {
   uchar depth = iExtPeriod;
   int index = startPos;
   //--- depth correction if need
   if ( (startPos-depth) < 0 )
      depth = (uchar)startPos;
   double max = CustPrevCalculated.GetDataHigh(startPos, false);
   //--- start searching
   double tmp;
   for (int i = startPos; i > (startPos-depth); i--)
     {
      tmp = CustPrevCalculated.GetDataHigh(i, false);
      if (tmp > max)
        {
         index = i;
         max = tmp;
        }
     }
   //--- return index of the highest bar
   return(index);
  }
//+------------------------------------------------------------------+
//|  searching index of the lowest bar                               |
//+------------------------------------------------------------------+
int CCustZigZagPPC::_iLowest(int startPos)
  {
   uchar depth = iExtPeriod;
   int index = startPos;
   //--- depth correction if need
   if ( (startPos-depth) < 0 )
      depth = (uchar)startPos;
   double min = CustPrevCalculated.GetDataLow(startPos, false);
   //--- start searching
   double tmp;
   for (int i = startPos; i > (startPos-depth); i--)
     {
      tmp = CustPrevCalculated.GetDataLow(i, false);
      if (tmp < min)
        {
         index = i;
         min = tmp;
        }
     }
   //--- return index of the lowest bar
   return(index);
  }
//+------------------------------------------------------------------+
//|                                     |
//+------------------------------------------------------------------+
char CCustZigZagPPC::_Comb(int i, double H, double L, double Fup, double Fdn)
{
   //----
   if (Fup==H && (Fdn==0 || (Fdn>0 && Fdn>L))) return(1);  //    
   if (Fdn==L && (Fup==0 || (Fup>0 && Fup<H))) return(-1); //    
   if (Fdn==L && Fup==H)                                   //        
     {
      OrderFormationBarHighLow OrderFormationHL = OFBError;
      //       OB[]
      if (Buf[PInd(i, false)].OB == OFBError)
        {
         OrderFormationHL = GetOrderFormationBarHighLow(Symbol(), Period(), DataBarsCount - i - 1, iUseSmallerTFforEB);
         if (OrderFormationHL != OFBError)
            Buf[PInd(i, false)].OB = OrderFormationHL;
         else
            return(CHAR_MIN); //          
        } else {
         OrderFormationHL = Buf[PInd(i, false)].OB;
        }
      
      switch(OrderFormationHL)
        {
         case OFBLowHigh:       //   LOW  HIGH ( )
            return(2);
            break;
         case OFBHighLow:       //   HIGH  LOW ( )
            return(-2); 
            break;
        }
     }
   //----  
   return(0);           //   ...
}
//+------------------------------------------------------------------+
CPCZZResultCode CCustZigZagPPC::Calculate()
{
   if (!Initialized)
      return(CPCZZRC_NotInitialized);
   
   //     ,   
   if (!UsesExternalData)
     {
      CPCPrepareDataResultCode resCPCPD = CustPrevCalculated.PrepareData();
      CPCPrepareDataResultCode resPD = PrepareData(resCPCPD);

      if (resPD == CPCPDRC_NoData)
         return(CPCZZRC_NoData);
      if (resPD == CPCPDRC_NoRecountNotRequired)
         return(CPCZZRC_NotChanged);
     }
   
   uint BarsForRecalculation;
   if ( DataBarsCalculated != 0 )
      BarsForRecalculation = DataBarsCount - ZZDataBarsCalculated - 1;
   else
      BarsForRecalculation = DataBarsCount - iExtPeriod - 1;
   
   //    
   if (ZZDataBarsCuttingLeft)
     {
      //    TimeFirstExtBar
      double tmpDN, tmpUP;
      uchar tmpC;
      for (uint i = 0; i < (DataBarsCount - BarsForRecalculation - 1); i++)
        {
         tmpC = 0;
         tmpDN = Buf[PInd(i, false)].DN;
         tmpUP = Buf[PInd(i, false)].UP;
         if (tmpDN != EMPTY_VALUE)
            tmpC = tmpC + 1;
         if (tmpUP != EMPTY_VALUE)
            tmpC = tmpC + 2;
         if ( tmpC == 3 )
           {
            Buf[PInd(i, false)].DN = EMPTY_VALUE;
            Buf[PInd(i, false)].UP = EMPTY_VALUE;
           }
         if ( (tmpC == 1) || (tmpC == 2) )
           { //    
            TimeFirstExtBar = CustPrevCalculated.GetDataTime(i, false);
            break;
           }
        }
      
      ZZDataBarsCuttingLeft = false;
     }
   
   //======================================================
   //========   ===============================
   //======================================================
   CPCZZResultCode res = CPCZZRC_NotChanged;
   
   char LET, tmpComb;
   double H, L, Fup, Fdn;
   uint lastUPbar, lastDNbar;
   double lastUP, lastDN;
   uint m, n; //    
   for (uint i = (DataBarsCount - BarsForRecalculation - 1); i < DataBarsCount; i++)
     {
      //   
      // ---------------------------
      lastUP = 0;
      lastDN = 0;
      lastUPbar = i;
      lastDNbar = i;
      LET = 0; 
      m = 0; n = 0;
      while ( Buf[PInd(lastUPbar, false)].UP == EMPTY_VALUE )
        {
         if ( lastUPbar < iExtPeriod )
            break;
         lastUPbar--;
        }
      lastUP = Buf[PInd(lastUPbar, false)].UP; //   
      while ( Buf[PInd(lastDNbar, false)].DN == EMPTY_VALUE)
        {
         if ( lastDNbar < iExtPeriod )
            break;
         lastDNbar--;
        } 
      lastDN = Buf[PInd(lastDNbar, false)].DN; //   
   
      if ( lastUPbar > lastDNbar)
         LET = 1;
      if ( lastUPbar < lastDNbar)
         LET = -1;
      if ( lastUPbar == lastDNbar)
        {
         //lastUPbar==lastDNbar  ,     :
         m = lastUPbar;
         n = m;
         while ( m == n )
           {
            m--; n--;
            while ( Buf[PInd(m, false)].UP == EMPTY_VALUE )
              {
               if ( m < iExtPeriod )
                  break;
               m--;
              } //   
            while ( Buf[PInd(n, false)].DN == EMPTY_VALUE )
              {
               if ( n < iExtPeriod )
                  break;
               n--;
              } //   
            if ( MathMax(m, n) < iExtPeriod )
               break;
           }
         if (m > n)
            LET = 1;       //  - 
         else
            if (m < n)
               LET = -1; //  - 
        }
      //  LET==0 -             2 
      //    
      // ----------------------------------

      //----      :       
      H = CustPrevCalculated.GetDataHigh(_iHighest(i), false);
      L = CustPrevCalculated.GetDataLow(_iLowest(i), false);
      Fup = CustPrevCalculated.GetDataHigh(i, false);
      Fdn = CustPrevCalculated.GetDataLow(i, false);
      
      //----        : 
      tmpComb = _Comb(i,H,L,Fup,Fdn);
      if (tmpComb == CHAR_MIN)
         return(CPCZZRC_NoData);
      switch(tmpComb)
        {
         //----      (Comb)      
         case 1 :
            switch(LET)
              {
               case 1 :
                  //   ,  :
                  if ( lastUP < Fup )
                    {
                     Buf[PInd(lastUPbar, false)].UP = EMPTY_VALUE;
                     Buf[PInd(i, false)].UP = Fup;
                     res = CPCZZRC_Changed;
                    }
                  break;
               case -1 :
                  if ( (Fup-lastDN) > iMinAmplitude ) //  - 
                    {
                     Buf[PInd(i, false)].UP = Fup;
                     res = CPCZZRC_Changed;
                    }
                  break; 
               default :
                  Buf[PInd(i, false)].UP = Fup;
                  TimeFirstExtBar = CustPrevCalculated.GetDataTime(i, false); //0 -     
                  res = CPCZZRC_Changed;
                  break; 
              }
            break;

         //----       (Comb)          
         case -1 :
            switch(LET)
              {
               case 1 :
                  if ( (lastUP-Fdn) > iMinAmplitude ) //  - 
                    {
                     Buf[PInd(i, false)].DN = Fdn;
                     res = CPCZZRC_Changed;
                    }
                  break; 
               case -1 :
                  //   ,  :
                  if ( lastDN > Fdn )
                    {
                     Buf[PInd(lastDNbar, false)].DN = EMPTY_VALUE;
                     Buf[PInd(i, false)].DN = Fdn;
                     res = CPCZZRC_Changed;
                    }
                  break;
               default :
                  Buf[PInd(i, false)].DN = Fdn;
                  TimeFirstExtBar = CustPrevCalculated.GetDataTime(i, false); //0 -     
                  res = CPCZZRC_Changed;
                  break;
              }
            break;

         //----         (Comb)        
         case 2 : //   LOW  HIGH ( )
            switch(LET)
              {
               case 1 : //  - 
                  if ( (Fup-Fdn) > iMinAmplitude )
                    {
                     if ( (lastUP-Fdn) > iMinAmplitude )
                       {
                        Buf[PInd(i, false)].UP = Fup;
                        Buf[PInd(i, false)].DN = Fdn;
                        res = CPCZZRC_Changed;
                       } else {
                        if ( lastUP < Fup )
                          {
                           Buf[PInd(lastUPbar, false)].UP = EMPTY_VALUE;
                           Buf[PInd(i, false)].UP = Fup;
                           res = CPCZZRC_Changed;
                          }
                       }
                    } else {
                     if ( (lastUP-Fdn) > iMinAmplitude )
                       {
                        Buf[PInd(i, false)].DN = Fdn;
                        res = CPCZZRC_Changed;
                       } else {
                        if ( lastUP < Fup )
                          {
                           Buf[PInd(lastUPbar, false)].UP = EMPTY_VALUE;
                           Buf[PInd(i, false)].UP = Fup;
                           res = CPCZZRC_Changed;
                          }
                       }
                    }
                  break;
               case -1 : //  - 
                  if ( (Fup-Fdn) > iMinAmplitude )
                    {
                     Buf[PInd(i, false)].UP = Fup;
                     if ( (Fdn < lastDN) && (CustPrevCalculated.GetDataTime(lastDNbar, false) > TimeFirstExtBar) )
                       {
                        Buf[PInd(lastDNbar, false)].DN = EMPTY_VALUE;
                        Buf[PInd(i, false)].DN = Fdn;
                       }
                     res = CPCZZRC_Changed;
                    } else {
                     if ( (Fup-lastDN) > iMinAmplitude )
                       {
                        Buf[PInd(i, false)].UP = Fup;
                        res = CPCZZRC_Changed;
                       } else {
                        if ( lastDN > Fdn )
                          {
                           Buf[PInd(lastDNbar, false)].DN = EMPTY_VALUE;
                           Buf[PInd(i, false)].DN = Fdn;
                           res = CPCZZRC_Changed;
                          }
                       }
                    }
                  break;
               default: break;
              } //switch LET
            break;

         //----         (Comb)        
         case -2 : //   HIGH  LOW ( )
            switch(LET)
              {
               case 1 : //  - 
                  if ( (Fup-Fdn) > iMinAmplitude )
                    {
                     Buf[PInd(i, false)].DN = Fdn;
                     if ( (lastUP < Fup) && (CustPrevCalculated.GetDataTime(lastUPbar, false) > TimeFirstExtBar) )
                       {
                        Buf[PInd(lastUPbar, false)].UP = EMPTY_VALUE;
                        Buf[PInd(i, false)].UP = Fup;
                       }
                     res = CPCZZRC_Changed;
                    } else {
                     if ( (lastUP-Fdn) > iMinAmplitude)
                       {
                        Buf[PInd(i, false)].DN = Fdn;
                        res = CPCZZRC_Changed;
                       } else {
                        if ( lastUP < Fup )
                          {
                           Buf[PInd(lastUPbar, false)].UP = EMPTY_VALUE;
                           Buf[PInd(i, false)].UP = Fup;
                           res = CPCZZRC_Changed;
                          }
                       }
                    }
                  break;
               case -1 : //  - 
                  if ( (Fup-Fdn) > iMinAmplitude )
                    {
                     if ( (Fup-lastDN) > iMinAmplitude )
                       {
                        Buf[PInd(i, false)].UP = Fup;
                        Buf[PInd(i, false)].DN = Fdn;
                        res = CPCZZRC_Changed;
                       } else {
                        if (lastDN > Fdn)
                          {
                           Buf[PInd(lastDNbar, false)].DN = EMPTY_VALUE;
                           Buf[PInd(i, false)].DN = Fdn;
                           res = CPCZZRC_Changed;
                          }
                       }
                    } else {
                     if ( (Fup-lastDN) > iMinAmplitude )
                       {
                        Buf[PInd(i, false)].UP = Fup;
                     res = CPCZZRC_Changed;
                       } else {
                        if ( lastDN > Fdn)
                          {
                           Buf[PInd(lastDNbar, false)].DN = EMPTY_VALUE;
                           Buf[PInd(i, false)].DN = Fdn;
                           res = CPCZZRC_Changed;
                          }
                       }
                    }
                  break;
               default: break;
              } //switch LET
            break;
         
         default: break;
        } // switch ()
      ZZDataBarsCalculated = i + 1; //  ,     
     } // for ()
   
   return(res);
}